unit UPascalParser;

{$mode delphi}

interface

uses
  Classes, SysUtils, UParser, USourceCode, Dialogs;

type
  TGetSourceEvent = function (Name: string; Source: TStringList): Boolean of object;

  { TPascalParser }

  TPascalParser = class(TBaseParser)
  private
    FOnGetSource: TGetSourceEvent;
  public
    function ParseFile(Name: string): Boolean;
    procedure ParseWhileDo(SourceCode: TWhileDo);
    procedure ParseExpression(SourceCode: TExpression);
    procedure ParseUses(SourceCode: TUsedModuleList; AExported: Boolean);
    function ParseModule(ProgramCode: TProgram): TModule;
    procedure ParseUnit(SourceCode: TModuleUnit);
    procedure ParseUnitInterface(SourceCode: TModuleUnit);
    procedure ParseUnitImplementation(SourceCode: TModuleUnit);
    procedure ParseProgram(SourceCode: TModuleProgram);
    procedure ParseCommonBlock(SourceCode: TCommonBlock; EndSymbol: char = ';';
      WithBody: Boolean = True);
    procedure ParseCommonBlockInterface(SourceCode: TCommonBlock);
    function ParseCommand(SourceCode: TCommonBlock): TCommand;
    procedure ParseBeginEnd(SourceCode: TBeginEnd);
    procedure ParseFunctionList(SourceCode: TFunctionList; Exported: Boolean = False);
    procedure ParseFunctionParameters(SourceCode: TFunction);
    procedure ParseIfThenElse(SourceCode: TIfThenElse);
    procedure ParseForToDo(SourceCode: TForToDo);
    procedure ParseVariableList(SourceCode: TVariableList; Exported: Boolean = False);
    procedure ParseVariable(SourceCode: TVariable; Exported: Boolean = False);
    procedure ParseConstantList(SourceCode: TConstantList; Exported: Boolean = False);
    procedure ParseTypeList(SourceCode: TTypeList; Exported: Boolean = False;
      AssignSymbol: string = '=');
    function ParseType(TypeList: TTypeList; ExpectName: Boolean = True; AssignSymbol: string = '='): TType;
    function ParseTypeSubType(TypeList: TTypeList; Name: string; ExpectName: Boolean): TType;
    function ParseTypeBase(TypeList: TTypeList; Name: string): TType;
    function ParseTypePointer(TypeList: TTypeList; Name: string): TType;
    function ParseTypeEnumeration(TypeList: TTypeList; Name: string): TType;
    function ParseTypeRecord(TypeList: TTypeList; Name: string): TType;
    function ParseTypeClass(TypeList: TTypeList; Name: string): TType;
    function ParseTypeArray(TypeList: TTypeList; Name: string): TType;
    function ParseTypeSubRange(TypeList: TTypeList; Name: string): TType;
    property OnGetSource: TGetSourceEvent read FOnGetSource
      write FOnGetSource;
    constructor Create;
    destructor Destroy; override;
  end;


resourcestring
  SUnknownIdentifier = 'Unknown identificator "%s".';
  SIllegalExpression = 'Illegal expression "%s".';
  SRedefineIdentifier = 'Identificator "%s" redefinition.';
  SEndOfDataReached = 'Parser reached to end of input data.';
  SUndefinedVariable = 'Undefined variable "%s".';
  SUndefinedType = 'Undefined type "%s".';
  SUndefinedConstant = 'Undefined constant "%s".';
  SUnitNotFound = 'Unit "%s" not found.';
  SFunctionNotDeclared = 'Function "%s" not declared.';
  SUnknownProcName = 'Unknown proc name "%s".';
  SUnknownModuleType = 'Unknown module name "%s".';
  SInvalidConstruction = 'Invalid construction.';

implementation

{ TPascalParser }

function TPascalParser.ParseFile(Name: string): Boolean;
var
  Parser: TPascalParser;
  NewModule: TModule;
begin
  try
    Parser := TPascalParser.Create;
    Parser.SourceCodeText := TStringList.Create;
    Parser.OnDebugLog := OnDebugLog;
    Parser.ProgramCode := ProgramCode;
    Parser.OnGetSource := OnGetSource;
    if Assigned(OnGetSource) then begin
      if FOnGetSource(Name, Parser.SourceCodeText) then begin
        Parser.Process;
        Parser.FileName := Name;
        Parser.OnErrorMessage := OnErrorMessage;
        //NewModule :=
        Parser.ParseModule(ProgramCode);
        //ProgramCode.Modules.Add(NewModule);
        Result := True;
      end else Result := False;
    end else Result := False;
  finally
    Parser.SourceCodeText.Free;
    Parser.Free;
  end;
end;

procedure TPascalParser.ParseWhileDo(SourceCode: TWhileDo);
begin
  with SourceCode do
  begin
    Expect('while');
    Condition.CommonBlock := CommonBlock;
    ParseExpression(Condition);
    Expect('do');
    Command := ParseCommand(CommonBlock);
  end;
end;

{ TExpression }

procedure TPascalParser.ParseExpression(SourceCode: TExpression);
var
  Identifier: string;
  IdentifierType: TTokenType;
  NewVariable: TVariable;
  NewExpression: TExpression;
  NewMethod: TFunction;
  Constant: TConstant;
  //  Brackets: Integer;
  Expressions: TExpressionList;
  I: integer;
  II: integer;
begin
  Expressions := TExpressionList.Create;
  Expressions.Add(TExpression.Create);
  with SourceCode do begin
    while ((NextToken <> ';') and (NextToken <> ',') and (not IsKeyWord(NextToken))) and not
      (((NextToken = ')') or (NextToken = ']'))) and not (NextTokenType = ttEndOfFile) do begin
      IdentifierType := NextTokenType;
      Identifier := ReadToken;
      if Identifier = '(' then begin
        // Subexpression
        with TExpression(Expressions.Last) do begin
          SubItems[1] := TExpression.Create;
          TExpression(SubItems[1]).CommonBlock := SourceCode.CommonBlock;
          ParseExpression(TExpression(SubItems[1]));
        end;
        with TExpression(Expressions.Items[Expressions.Add(TExpression.Create)]) do
        begin
          CommonBlock := SourceCode.CommonBlock;
          SubItems[0] := TExpression(Expressions[Expressions.Count - 2]).SubItems[1];
        end;
        Expect(')');
      end else
      if IsOperator(Identifier) then begin
        // Operator
        TExpression(Expressions.Last).OperatorName := Identifier;
        TExpression(Expressions.Last).NodeType := ntOperator;
      end else
      if IsIdentificator(Identifier) then begin
        // Reference to identificator
        NewVariable := CommonBlock.Variables.Search(Identifier);
        if Assigned(NewVariable) then begin
          // Referenced variable
          with TExpression(Expressions.Last) do begin
            SubItems[1] := TExpression.Create;
            TExpression(SubItems[1]).CommonBlock := SourceCode.CommonBlock;
            TExpression(SubItems[1]).NodeType := ntVariable;
            TExpression(SubItems[1]).Variable := NewVariable;
          end;
          with TExpression(Expressions.Items[Expressions.Add(TExpression.Create)]) do
          begin
            CommonBlock := SourceCode.CommonBlock;
            SubItems[0] := TExpression(Expressions[Expressions.Count - 2]).SubItems[1];
          end;
        end else begin
          NewMethod := CommonBlock.Functions.Search(Identifier);
          if Assigned(NewMethod) then begin
            // Referenced method
            with TExpression(Expressions.Last) do begin
              SubItems[1] := TExpression.Create;
              TExpression(SubItems[1]).CommonBlock := SourceCode.CommonBlock;
              if NextToken = '(' then               // Method with parameters
                with TExpression(SubItems[1]) do begin
                  Expect('(');
                  NewExpression := TExpression.Create;
                  NewExpression.CommonBlock := CommonBlock;
                  ParseExpression(NewExpression);
                  SubItems.Add(NewExpression);
                  while NextToken = ',' do begin
                    Expect(',');
                    NewExpression := TExpression.Create;
                    NewExpression.CommonBlock := CommonBlock;
                    ParseExpression(NewExpression);
                    SubItems.Add(NewExpression);
                  end;
                  Expect(')');
                end;
              TExpression(SubItems[1]).NodeType := ntFunction;
              TExpression(SubItems[1]).Method := NewMethod;
            end;
            with TExpression(Expressions.Items[Expressions.Add(TExpression.Create)]) do
            begin
              CommonBlock := SourceCode.CommonBlock;
              SubItems[0] :=
                TExpression(Expressions[Expressions.Count - 2]).SubItems[1];
            end;
          end else begin
            Constant := CommonBlock.Constants.Search(Identifier);
            if Assigned(Constant) then begin
              // Referenced constant
              with TExpression(Expressions.Last) do begin
                SubItems[1] := TExpression.Create;
                TExpression(SubItems[1]).CommonBlock := SourceCode.CommonBlock;
                TExpression(SubItems[1]).NodeType := ntConstant;
                TExpression(SubItems[1]).Value := Constant.Value;
              end;
              with TExpression(Expressions.Items[Expressions.Add(
                  TExpression.Create)]) do begin
                CommonBlock := SourceCode.CommonBlock;
                SubItems[0] :=
                  TExpression(Expressions[Expressions.Count - 2]).SubItems[1];
              end;
            end else begin
              ErrorMessage(SUnknownIdentifier, [Identifier], -1);
            end;
          end;
        end;
      end else begin
        // Constant value
        with TExpression(Expressions.Last) do begin
          SubItems[1] := TExpression.Create;
          TExpression(SubItems[1]).CommonBlock := SourceCode.CommonBlock;
          TExpression(SubItems[1]).NodeType := ntConstant;

          if IdentifierType = ttConstantString then begin
            TExpression(SubItems[1]).Value := Identifier;
            //SetLength(TExpression(SubItems[1]).Value, Length(Identifier));
            //for I := 1 to Length(Identifier) do
            //  TExpression(SubItems[1]).Value[I - 1] := Byte(Identifier[I]);
          end else begin
            TExpression(SubItems[1]).Value := Identifier;
          end;
        end;
        //ShowMessage(IntToStr(Expressions.Count));
        with TExpression(Expressions.Items[Expressions.Add(TExpression.Create)]) do
        begin
          CommonBlock := SourceCode.CommonBlock;
          SubItems[0] := TExpression(Expressions[Expressions.Count - 2]).SubItems[1];
        end;
      end;
    end;

    // Build expression tree
    for II := 0 to High(Operators) do begin
      I := 1;
      while (I < Expressions.Count - 1) do begin
        if not TExpression(Expressions[I]).Associated and
          (TExpression(Expressions[I]).OperatorName = Operators[II]) then
        begin
          TExpression(Expressions[I]).Associated := True;
          TExpression(Expressions[I - 1]).SubItems[1] := Expressions[I];
          TExpression(Expressions[I + 1]).SubItems[0] := Expressions[I];
          //Expressions.Delete(I);
        end else Inc(I);
      end;
    end;
    if Assigned(TExpression(Expressions.First).SubItems[1]) then
      Assign(TExpression(TExpression(Expressions.First).SubItems[1]));
    TExpression(Expressions.First).SubItems[1] := nil;
    //ShowMessage(IntToStr(Expressions.Count));
    if Expressions.Count > 1 then
      TExpression(Expressions[1]).SubItems[0] := nil;
    Expressions.Free;
  end;
end;

function TPascalParser.ParseCommand(SourceCode: TCommonBlock): TCommand;
var
  Identifier: string;
  Variable: TVariable;
  Method: TFunction;
  First: TOperation;
  Second: TOperation;
  StartIndex: integer;
  LoopVariable: TVariable;
  IdentName: string;
  FunctionName: string;
begin
  begin
    if NextToken = 'begin' then begin
      Result := TBeginEnd.Create;
      TBeginEnd(Result).CommonBlock := SourceCode;
      //ShowMessage(IntToStr(Integer(SourceCode))
      // + ' ' + IntToStr(Integer(Result)));
      ParseBeginEnd(TBeginEnd(Result));
    end else
    if NextToken = 'if' then begin
      Result := TIfThenElse.Create;
      TIfThenElse(Result).CommonBlock := SourceCode;
      ParseIfThenElse(TIfThenElse(Result));
    end else
    if NextToken = 'while' then begin
      Result := TWhileDo.Create;
      TWhileDo(Result).CommonBlock := SourceCode;
      ParseWhileDo(TWhileDo(Result));
    end else
    if NextToken = 'for' then begin
      Result := TForToDo.Create;
      TForToDo(Result).CommonBlock := SourceCode;
      ParseForToDo(TForToDo(Result));
    end else
    if IsIdentificator(NextToken) then begin
      if Assigned(SourceCode.Variables.Search(NextToken)) then begin
        // Variable assignment
        Result := TAssignment.Create;
        TAssignment(Result).CommonBlock := SourceCode;
        IdentName := ReadToken;
        TAssignment(Result).Target := SourceCode.Variables.Search(IdentName);
        Expect(':=');
        TAssignment(Result).Source := TExpression.Create;
        TAssignment(Result).Source.CommonBlock := SourceCode;
        ParseExpression(TAssignment(Result).Source);
      end else
      if Assigned(SourceCode.Functions.Search(NextToken)) then begin
        // Function call
        FunctionName := ReadToken;
        Result := TFunctionCall.Create;
        TFunctionCall(Result).CommonBlock := SourceCode;
        TFunctionCall(Result).FunctionRef := SourceCode.Functions.Search(FunctionName);
        if NextToken = '(' then
        begin
          Expect('(');
          with TFunctionCall(Result) do
          begin
            ParameterExpression.Add(TExpression.Create);
            TExpression(ParameterExpression.Last).CommonBlock := SourceCode;
            ParseExpression(TExpression(ParameterExpression.Last));
          end;
          Expect(')');
        end;
      end else begin
        Result := nil;
        ErrorMessage(SUnknownIdentifier, [ReadToken], -1);
      end;
    end else
    if NextToken = ';' then
      Result := nil
    else begin
      Result := nil;
      ErrorMessage(SIllegalExpression, [ReadToken], -1);
    end;
  end;
end;

{ TParserModule }

function TPascalParser.ParseModule(ProgramCode: TProgram): TModule;
begin
  Self.ProgramCode := ProgramCode;
  if NextToken = 'unit' then begin
    Result := TModuleUnit.Create;
    Result.ParentProgram := ProgramCode;
    ParseUnit(TModuleUnit(Result));
  end else
  if NextToken = 'program' then begin
    Result := TModuleProgram.Create;
    Result.ParentProgram := ProgramCode;
    ParseProgram(TModuleProgram(Result));
  end else
    ErrorMessage(SUnknownModuleType, [NextToken]);
end;

procedure TPascalParser.ParseProgram(SourceCode: TModuleProgram);
var
  Identifier: string;
begin
  with SourceCode do begin
    if NextToken = 'program' then begin
      Expect('program');
      Name := ReadToken;
      Expect(';');
    end else Name := '';

    // Uses section
    if NextToken = 'uses' then
      ParseUses(UsedModules, False);

    ParseCommonBlock(Body, '.');
    SourceCode.ParentProgram.Modules.Add(SourceCode);
  end;
end;

procedure TPascalParser.ParseUnit(SourceCode: TModuleUnit);
var
  NewModule: TModule;
begin
  Expect('unit');
  with Sourcecode do begin
    Name := ReadToken;
  end;
  Expect(';');

  ParseUnitInterface(SourceCode);
  if NextToken = 'implementation' then
    ParseUnitImplementation(SourceCode);

  SourceCode.ParentProgram.Modules.Add(SourceCode);

  if NextToken = 'initialization' then begin
    Expect('initialization');
  end;
  if NextToken = 'finalization' then begin
    Expect('finalization');
  end;
end;

procedure TPascalParser.ParseUnitInterface(SourceCode: TModuleUnit);
begin
  Expect('interface');
  // Uses section
  if NextToken = 'uses' then
    ParseUses(SourceCode.UsedModules, True);

  ParseCommonBlockInterface(SourceCode.Body);
end;

procedure TPascalParser.ParseUnitImplementation(SourceCode: TModuleUnit);
begin
  Expect('implementation');

  // Uses section
  if NextToken = 'uses' then
    ParseUses(SourceCode.UsedModules, False);

  ParseCommonBlock(SourceCode.Body, '.', False);
end;

{ TParserCommonBlock }

procedure TPascalParser.ParseCommonBlock(SourceCode: TCommonBlock;
  EndSymbol: char = ';'; WithBody: Boolean = True);
begin
  with SourceCode do begin
    while (NextToken <> EndSymbol) do begin
      if NextToken = 'var' then begin
        Expect('var');
        ParseVariableList(Variables)
      end else
      if NextToken = 'const' then begin
        Expect('const');
        ParseConstantList(Constants)
      end else
      if NextToken = 'type' then begin
        Expect('type');
        ParseTypeList(Types);
      end else
      if NextToken = 'procedure' then
        ParseFunctionList(Functions)
      else if NextToken = 'function' then
        ParseFunctionList(Functions)
      else begin
        if WithBody then
          ParseBeginEnd(Code);
        Break;
      end;
    end;
    if WithBody then Expect(EndSymbol);
  end;
end;

procedure TPascalParser.ParseCommonBlockInterface(SourceCode: TCommonBlock);
begin
  with SourceCode do begin
    while (NextToken <> 'implementation') and (NextTokenType <> ttEndOfFile) do begin
      if NextToken = 'var' then begin
        Expect('var');
        ParseVariableList(Variables);
      end else
      if NextToken = 'const' then begin
        Expect('const');
        ParseConstantList(Constants, True);
      end else
      if NextToken = 'type' then begin
        Expect('type');
        ParseTypeList(Types, True);
      end else
      if NextToken = 'procedure' then
        ParseFunctionList(Functions, True)
      else if NextToken = 'function' then
        ParseFunctionList(Functions, True)
      else begin
        ErrorMessage(SUnknownIdentifier, [NextToken], -1);
        ReadToken;
      end;
    end;
  end;
end;

{ TParserBeginEnd }

procedure TPascalParser.ParseBeginEnd(SourceCode: TBeginEnd);
var
  NewCommand: TCommand;
begin
  //ShowMessage(IntToStr(Integer(SourceCode)) + ' ' + IntToStr(Integer(SourceCode.CommonBlock)));
  with SourceCode do
  begin
    Expect('begin');
    while (NextToken <> 'end') and (NextTokenType <> ttEndOfFile) do
    begin
      NewCommand := ParseCommand(CommonBlock);
      if Assigned(NewCommand) then
        Commands.Add(NewCommand);
      //ShowMessage(NextCode);
      if NextToken = ';' then
        ReadToken;
    end;
    Expect('end');
  end;
end;

{ TParserParseFunctionList }

procedure TPascalParser.ParseFunctionList(SourceCode: TFunctionList;
  Exported: Boolean = False);
var
  NewValueType: TType;
  TypeName: string;
  UseName: string;
  I: Integer;
  UseType: TType;
  UseFunction: TFunction;
  FunctionType: TFunctionType;
begin
  with SourceCode do begin
    if NextToken = 'procedure' then begin
      Expect('procedure');
      FunctionType := ftProcedure;
    end else
    if NextToken = 'function' then begin
      Expect('function');
      FunctionType := ftFunction;
    end else
    if NextToken = 'constructor' then begin
      Expect('constructor');
      FunctionType := ftConstructor;
    end else
    if NextToken = 'destructor' then begin
      Expect('destructor');
      FunctionType := ftDestructor;
    end else ErrorMessage(SUnknownProcName, [NextToken]);

    // Read function name
    UseName := ReadToken;
    UseType := SourceCode.Parent.Types.Search(UseName);
    if Assigned(UseType) and ((UseType is TTypeRecord) or
    (UseType is TTypeClass)) then begin
      Expect('.');
      UseName := ReadToken;
      if UseType is TTypeRecord then begin
        UseFunction := TTypeRecord(UseType).CommonBlock.Functions.Search(UseName);
        if not Assigned(UseFunction) then begin
          ErrorMessage(SFunctionNotDeclared, [UseName]);
          Exit;
        end;
      end;
    end else begin
      // Create new function
      UseFunction := TFunction.Create;
      UseFunction.Parent := SourceCode.Parent;
      UseFunction.Name := UseName;
      UseFunction.FunctionType := FunctionType;
      Add(UseFunction);
    end;
    with UseFunction do begin
      // Parse parameters
      if NextToken = '(' then
        ParseFunctionParameters(UseFunction);

      // Parse function result type
      if FunctionType = ftFunction then begin
        Expect(':');
        TypeName := ReadToken;
        NewValueType := Parent.Types.Search(TypeName);
        if not Assigned(NewValueType) then
          ErrorMessage(SUndefinedType, [TypeName], -1);
(*        else
          begin
            ResultType := NewValueType;
            with TVariable(Parent.Variables.Items[Parent.Variables.Add(
                TVariable.Create)]) do
            begin
              Name := 'Result';
              ValueType := NewValueType;
            end;
          end;  *)
      end;
      Expect(';');

      // Check directives
      if NextToken = 'internal' then begin
        Expect('internal');
        Expect(';');
        Internal := True;
      end;
    end;

    if not Exported then ParseCommonBlock(UseFunction);
//    if UseFunction then UseFunction.Code ;
  end;
end;

procedure TPascalParser.ParseFunctionParameters(SourceCode: TFunction);
var
  Identifiers: TStringList;
  VariableName: string;
  UseVariable: TParameter;
  TypeName: string;
  UseType: TType;
  I: Integer;
begin
  with SourceCode do
  try
    Identifiers := TStringList.Create;
        Expect('(');
        while NextToken <> ')' do begin
          // while IsIdentificator(NextCode) do begin
          with TParameterList(Parameters) do begin
            VariableName := ReadToken;
            if VariableName = 'var' then begin
            end else
            if VariableName = 'const' then begin
            end else begin
              UseVariable := Search(VariableName);
              if not Assigned(UseVariable) then begin
                Identifiers.Add(VariableName);
                while NextToken = ',' do begin
                  Expect(',');
                  Identifiers.Add(ReadToken);
                end;
              end else
                ErrorMessage(SRedefineIdentifier, [VariableName], -1);
              Expect(':');
              TypeName := ReadToken;
              UseType := Parent.Types.Search(TypeName);
              if not Assigned(UseType) then
                ErrorMessage(SUndefinedType, [TypeName], -1)
              else
                for I := 0 to Identifiers.Count - 1 do
                  with TParameter(Items[Add(TParameter.Create)]) do
                  begin
                    Name := Identifiers[I];
                    ValueType := UseType;
                  end;
            end;
          end;
          if NextToken = ';' then Expect(';');
        end;
        Expect(')');
  finally
    Identifiers.Free;
  end;
end;

{ TParserIfThenElse }

procedure TPascalParser.ParseIfThenElse(SourceCode: TIfThenElse);
begin
  with Sourcecode do
  begin
    Expect('if');
    Condition.CommonBlock := CommonBlock;
    ParseExpression(Condition);
    Expect('then');
    Command := ParseCommand(CommonBlock);
    if NextToken = 'else' then
    begin
      Expect('else');
      ElseCommand := ParseCommand(CommonBlock);
    end;
  end;
end;

procedure TPascalParser.ParseForToDo(SourceCode: TForToDo);
var
  VariableName: string;
begin
  with SourceCode do
  begin
    Expect('for');
    VariableName := ReadToken;
    ControlVariable := SourceCode.CommonBlock.Variables.Search(VariableName);
    if not Assigned(ControlVariable) then
      ErrorMessage(SUndefinedVariable, [VariableName], -1);
    Expect(':=');
    Start.CommonBlock := CommonBlock;
    ParseExpression(Start);
    Expect('to');
    Stop.CommonBlock := CommonBlock;
    ParseExpression(Stop);
    Expect('do');
    Command := ParseCommand(CommonBlock);
  end;
end;

{ TParserVariableList }

procedure TPascalParser.ParseVariableList(SourceCode: TVariableList; Exported: Boolean = False);
var
  Identifiers: TStringList;
  NewValueType: TType;
  TypeName: string;
  VariableName: string;
  Variable: TVariable;
  I: integer;
begin
  try
  Identifiers := TStringList.Create;
  with SourceCode do begin
    while IsIdentificator(NextToken) and (NextTokenType <> ttEndOfFile) do begin
      Identifiers.Clear;
      VariableName := ReadToken;
      Variable := Search(VariableName);
      if not Assigned(Variable) then begin
        Identifiers.Add(VariableName);
        while NextToken = ',' do begin
          Expect(',');
          Identifiers.Add(ReadToken);
        end;
      end else
        ErrorMessage(SRedefineIdentifier, [VariableName], -1);
      Expect(':');
      TypeName := ReadToken;
      NewValueType := Parent.Types.Search(TypeName);
      if NewValueType = nil then
        ErrorMessage(SUndefinedType, [TypeName], -1)
      else
        for I := 0 to Identifiers.Count - 1 do
          with TVariable(Items[Add(TVariable.Create)]) do begin
            Name := Identifiers[I];
            ValueType := NewValueType;
          end;
      Expect(';');
    end;
  end;
  finally
    Identifiers.Free;
  end;
end;

{ TParserVariable }

procedure TPascalParser.ParseVariable(SourceCode: TVariable; Exported: Boolean = False);
begin
  with SourceCode do begin
    Name := NextToken;
    Expect(':=');

  end;
end;

{ TParserConstantList }

procedure TPascalParser.ParseConstantList(SourceCode: TConstantList; Exported: Boolean = False);
var
  Identifiers: TStringList;
  NewValueType: TType;
  TypeName: string;
  ConstantName: string;
  Constant: TConstant;
  I: integer;
  ConstantValue: string;
begin
  Identifiers := TStringList.Create;
  with SourceCode do begin
    while IsIdentificator(NextToken) do begin
      ConstantName := ReadToken;
      Constant := Search(ConstantName);
      if not Assigned(Constant) then begin
        Identifiers.Add(ConstantName);
        while NextToken = ',' do begin
          Expect(',');
          Identifiers.Add(ReadToken);
        end;
      end else
        ErrorMessage(SRedefineIdentifier, [ConstantName], -1);
      if NextToken = ':' then begin
        Expect(':');
        TypeName := ReadToken;
        NewValueType := Parent.Types.Search(TypeName);
      end;
      Expect('=');
      ConstantValue := ReadToken;
      Expect(';');

      if NewValueType = nil then
        ErrorMessage(SUndefinedType, [TypeName], -1)
      else
        for I := 0 to Identifiers.Count - 1 do
          with TConstant(Items[Add(TConstant.Create)]) do
          begin
            Name := Identifiers[I];
            ValueType := NewValueType;
            Value := ConstantValue;
          end;
    end;
  end;
  Identifiers.Destroy;
end;

{ TParserTypeList }

procedure TPascalParser.ParseTypeList(SourceCode: TTypeList;
  Exported: Boolean = False; AssignSymbol: string = '=');
var
  NewType: TType;
begin
  with SourceCode do
  begin
    while IsIdentificator(NextToken) do begin
      NewType := ParseType(SourceCode, True, AssignSymbol);
      if Assigned(NewType) then begin
        NewType.Parent := SourceCode;
        Add(NewType);
      end;
      Expect(';');
    end;
  end;
end;

{ TParserType }

function TPascalParser.ParseType(TypeList: TTypeList; ExpectName: Boolean = True;
  AssignSymbol: string = '='): TType;
var
  Name: string;
  TypeName: string;
begin
  //with SourceCode do
  begin
    if ExpectName then begin
      Name := ReadToken;
      Expect(AssignSymbol);
    end;
    Result := ParseTypeEnumeration(TypeList, Name);
    if not Assigned(Result) then Result := ParseTypeRecord(TypeList, Name);
    if not Assigned(Result) then Result := ParseTypeClass(TypeList, Name);
    if not Assigned(Result) then Result := ParseTypeArray(TypeList, Name);
    if not Assigned(Result) then Result := ParseTypePointer(TypeList, Name);
    if not Assigned(Result) then Result := ParseTypeBase(TypeList, Name);
    if not Assigned(Result) then Result := ParseTypeSubType(TypeList, Name, ExpectName);
    if not Assigned(Result) then Result := ParseTypeSubRange(TypeList, Name);
    if not Assigned(Result) then
      ErrorMessage(SInvalidConstruction, []);
  end;
end;

function TPascalParser.ParseTypeSubType(TypeList: TTypeList; Name: string;
  ExpectName: Boolean): TType;
var
  TypeName: string;
begin
      // Use existed type
      if NextTokenType = ttIdentifier then begin
        TypeName := ReadToken;
        if ExpectName then begin
          Result := TType.Create;
          TType(Result).Parent := TypeList;
          TType(Result).Name := Name;
          TType(Result).UsedType := TypeList.Search(TypeName);
          if not Assigned(TType(Result).UsedType) then
            ErrorMessage(SUndefinedType, [TypeName], -1);
        end else begin
          TType(Result) := TypeList.Search(TypeName);
          if not Assigned(TType(Result)) then
            ErrorMessage(SUndefinedType, [TypeName], -1);
        end;
      end else Result := nil;
end;

function TPascalParser.ParseTypeBase(TypeList: TTypeList; Name: string): TType;
begin
      // Buildin base type construction
      if NextToken = 'type' then begin
        Expect('type');
        Result := TTypeInherited.Create;
        TTypeInherited(Result).Parent := TypeList;
        TTypeInherited(Result).Name := Name;
        if NextToken = '(' then begin
          Expect('(');
          TTypeInherited(Result).UsedType := ParseType(TypeList, False);
          Expect(')');
        end else TTypeInherited(Result).UsedType := nil;
      end else Result := nil;
end;

function TPascalParser.ParseTypePointer(TypeList: TTypeList; Name: string
  ): TType;
begin
      if NextToken = '^' then begin
        Expect('^');
        Result := TTypePointer.Create;
        TTypePointer(Result).Parent := TypeList;
        TTypePointer(Result).Name := Name;
        TTypePointer(Result).UsedType := ParseType(TypeList, False);
      end else Result := nil;
end;

function TPascalParser.ParseTypeEnumeration(TypeList: TTypeList; Name: string): TType;
begin
  if NextToken = '(' then begin
      Expect('(');
      Result := TTypeEnumeration.Create;
      TTypeEnumeration(Result).Parent := TypeList;
      TTypeEnumeration(Result).Name := Name;
      with TTypeEnumeration(Result) do
      with TEnumItem(Items[Items.Add(TEnumItem.Create)]) do begin
        Name := ReadToken;
        if (NextToken = '=') and (NextTokenType = ttConstantNumber) then begin
          Expect('=');
          Index := StrToInt(ReadToken);
        end;
      end;
      while (NextToken = ',') and (NextTokenType <> ttEndOfFile) do
      begin
        Expect(',');
        with TTypeEnumeration(Result) do
        with TEnumItem(Items[Items.Add(TEnumItem.Create)]) do begin
          Name := ReadToken;
          if (NextToken = '=') and (NextTokenType = ttConstantNumber) then begin
            Expect('=');
            Index := StrToInt(ReadToken);
          end;
        end;
      end;
      Expect(')');
  end else Result := nil;
end;

function TPascalParser.ParseTypeRecord(TypeList: TTypeList; Name: string
  ): TType;
type
  TSectionType = (stVar, stType, stConst);
var
  Visibility: TTypeVisibility;
  SectionType: TSectionType;
begin
  if NextToken = 'record' then begin
  SectionType := stVar;
  Visibility := tvPublic;
  Expect('record');
  Result := TTypeRecord.Create;
  TTypeRecord(Result).Parent := TypeList;
  TTypeRecord(Result).CommonBlock.Parent := TypeList.Parent;
  TType(Result).Name := Name;
  while (NextToken <> 'end') and (NextTokenType <> ttEndOfFile) do begin
    if NextToken = 'public' then begin
      Expect('public');
      Visibility := tvPublic;
    end else
    if NextToken = 'private' then begin
      Expect('private');
      Visibility := tvPrivate;
    end else
    if NextToken = 'published' then begin
      Expect('published');
      Visibility := tvPublished;
    end else
    if NextToken = 'protected' then begin
      Expect('protected');
      Visibility := tvProtected;
    end else
    if NextToken = 'var' then begin
      Expect('var');
      SectionType := stVar;
    end else
    if NextToken = 'const' then begin
      Expect('const');
      SectionType := stConst;
    end else
    if NextToken = 'type' then begin
      Expect('type');
      SectionType := stType;
    end else
    if NextToken = 'procedure' then
      ParseFunctionList(TTypeRecord(Result).CommonBlock.Functions, True)
    else if NextToken = 'function' then
      ParseFunctionList(TTypeRecord(Result).CommonBlock.Functions, True)
    else begin
      if SectionType = stVar then begin
          if IsIdentificator(NextToken) then
            ParseVariableList(TTypeRecord(Result).CommonBlock.Variables, True)
            else ReadToken;
          //TTypeRecord(Result).CommonBlock.Types.Add(ParseType(TypeList, True, ':'));
          //TType(TTypeRecord(Result).CommonBlock.Types.Last).Visibility := Visibility;
      end
      else if SectionType = stConst then
        ParseConstantList(TTypeRecord(Result).CommonBlock.Constants, True)
      else if SectionType = stType then
        ParseTypeList(TTypeRecord(Result).CommonBlock.Types, True, '=');
    end;
  end;
  Expect('end');
  end else Result := nil;
end;

function TPascalParser.ParseTypeClass(TypeList: TTypeList; Name: string
  ): TType;
begin
  if NextToken = 'class' then begin
      Expect('class');
      Result := TTypeClass.Create;
      TTypeClass(Result).Parent := TypeList;
      TTypeClass(Result).Name := Name;
      if NextToken <> ';' then begin
        while (NextToken <> 'end') and (NextTokenType <> ttEndOfFile) do
        begin
          TTypeClass(Result).Items.Add(ParseType(TypeList, True, ':'));
          Expect(';');
        end;
        Expect('end');
      end;
  end else Result := nil;
end;

function TPascalParser.ParseTypeArray(TypeList: TTypeList; Name: string
  ): TType;
var
  UseName: string;
  UseType: TType;
begin
  if NextToken = 'array' then begin
  Expect('array');
  Result := TTypeArray.Create;
  TTypeArray(Result).Parent := TypeList;
  TType(Result).Name := Name;
  if NextToken = '[' then begin
    Expect('[');
    UseName := NextToken;
    if NextTokenType = ttIdentifier then begin
      UseType := TypeList.Parent.Types.Search(UseName);
      if not Assigned(TTypeArray(Result).IndexType) then
        ErrorMessage(SUndefinedType, [UseName], -1) else
        TTypeArray(Result).IndexType := UseType;
    end else
    if NextTokenType = ttConstantNumber then begin


    end;
    Expect(']');
  end;
  Expect('of');
  UseName := NextToken;
  TTypeArray(Result).ItemType := ParseType(TypeList, False);
  if not Assigned(TTypeArray(Result).ItemType) then
    ErrorMessage(SUndefinedType, [UseName], -1);
  end else Result := nil;
end;

function TPascalParser.ParseTypeSubRange(TypeList: TTypeList; Name: string
  ): TType;
var
  UseName: string;
begin
  if NextTokenType = ttConstantString then begin
    Result := TTypeSubRange.Create;
    TTypeSubRange(Result).Bottom := ReadToken;
    Expect('..');
    TTypeSubRange(Result).Top := ReadToken;
  end else
  if NextTokenType = ttConstantNumber then begin
    Result := TTypeSubRange.Create;
    TTypeSubRange(Result).Bottom := ReadToken;
    Expect('..');
    TTypeSubRange(Result).Top := ReadToken;
  end else Result := nil;
end;

constructor TPascalParser.Create;
begin
  inherited;
end;

destructor TPascalParser.Destroy;
begin
  inherited Destroy;
end;

{ TParserUsedModuleList }

procedure TPascalParser.ParseUses(SourceCode: TUsedModuleList; AExported: Boolean = False);
var
  NewUsedModule: TUsedModule;
begin
  Expect('uses');
  with TUsedModule(SourceCode.Items[SourceCode.Add(TUsedModule.Create)]) do
  begin
    Name := ReadToken;
    if NextToken = 'in' then begin
      Expect('in');
      Location := ReadToken;
    end else Location := Name + '.pas';
    Module := SourceCode.ParentModule.ParentProgram.Modules.Search(Name);
    if not Assigned(Module) then begin
      if ParseFile(Name) then begin
        Module := SourceCode.ParentModule.ParentProgram.Modules.Search(Name);
        Exported := AExported;
      end else begin
        ErrorMessage(SUnitNotFound, [Name], -2);
        SourceCode.Delete(SourceCode.Count - 1);
      end;
    end;
  end;
  while NextToken = ',' do begin
    Expect(',');
    with TUsedModule(SourceCode.Items[SourceCode.Add(TUsedModule.Create)]) do
    begin
      Name := ReadToken;
      if NextToken = 'in' then begin
        Expect('in');
        Location := ReadToken;
      end else Location := Name + '.pas';
      Module := SourceCode.ParentModule.ParentProgram.Modules.Search(Name);
      if not Assigned(Module) then begin
        if not ParseFile(Name) then begin
          ErrorMessage(SUnitNotFound, [Name], -2);
          SourceCode.Delete(SourceCode.Count - 1);
        end;
      end;
    end;
  end;
  Expect(';');
end;


end.

